"""General tools which don't depend on other parts of Parsimonious"""

import ast


class StrAndRepr(object):
    """Mix-in which gives the class the same __repr__ and __str__."""

    def __repr__(self):
        return self.__str__()


def evaluate_string(string):
    """Piggyback on Python's string support so we can have backslash escaping
    and niceties like \n, \t, etc. string.decode('string_escape') would have
    been a lower-level possibility.

    """
    return ast.literal_eval(string)


class Token(StrAndRepr):
    """A class to represent tokens, for use with TokenGrammars

    You will likely want to subclass this to hold additional information, like
    the characters that you lexed to create this token. Alternately, feel free
    to create your own class from scratch. The only contract is that tokens
    must have a ``type`` attr.

    """
    __slots__ = ['type']

    def __init__(self, type):
        self.type = type

    def __str__(self):
        return u'<Token "%s">' % (self.type,)

    def __eq__(self, other):
        return self.type == other.type



import asyncio
from ccxt import ExchangeClosedByUser

class Future(asyncio.Future):

    is_race_future = False

    def resolve(self, result=None):
        if not self.done():
            try:
                self.set_result(result)
            except BaseException as e:
                print("Error in Future.resolve")
                raise e

    def reject(self, error=None):
        if not self.done():
            # If not an exception, wrap it in a generic Exception
            if not isinstance(error, BaseException):
                error = Exception(error)
            try:
                self.set_exception(error)
            except BaseException as e:
                print("Error in Future.reject")
                raise e

    @classmethod
    def race(cls, futures):
        future = Future()
        for f in futures:
            f.is_race_future = True
        coro = asyncio.wait(futures, return_when=asyncio.FIRST_COMPLETED)
        task = asyncio.create_task(coro)

        def callback(done):
            try:
                complete, pending = done.result()
                # check for exceptions
                for i, f in enumerate(complete):
                    try:
                        f.result()
                    except ExchangeClosedByUser as e:
                        if len(pending) == 0 and i == len(complete) - 1:
                            future.reject(e)
                        # wait for all the sub promises to be reject before rejecting future
                        continue
                    except asyncio.CancelledError as e:
                        continue
                    except Exception as e:
                        future.reject(e)
                        return
                # no exceptions return first result
                futures_list = list(complete)

                are_all_canceled = all([f.cancelled() for f in futures_list])
                if are_all_canceled:
                    future.reject(ExchangeClosedByUser('Connection closed by the user'))
                    return

                first = futures_list[0]

                first_result = first.result()
                future.resolve(first_result)
            except asyncio.CancelledError as e:
                future.reject(e)
            except Exception as e:
                future.reject(e)
        task.add_done_callback(callback)
        return future



import typing

import marshmallow


class Sequence(marshmallow.fields.List):
    """
    A sequence field, basically an immutable version of the list field.
    """

    def _deserialize(  # type: ignore[override]
        self,
        value: typing.Any,
        attr: typing.Any,
        data: typing.Any,
        **kwargs: typing.Any,
    ) -> typing.Optional[typing.Sequence[typing.Any]]:
        optional_list = super()._deserialize(value, attr, data, **kwargs)
        return None if optional_list is None else tuple(optional_list)


class Set(marshmallow.fields.List):
    """
    A set field. A set is an unordered/mutable collection of unique elements, same for frozenset
    except it's immutable.

    Notes:
        Beware the a Set guarantees uniqueness in the resulting list but in return the item's order
        will be random. So if the order matters, use a List or Sequence !
    """

    def __init__(
        self,
        cls_or_instance: typing.Union[marshmallow.fields.Field, type],
        frozen: bool = False,
        **kwargs,
    ):
        super().__init__(cls_or_instance, **kwargs)
        self.set_type: typing.Type[typing.Union[frozenset, set]] = (
            frozenset if frozen else set
        )

    def _deserialize(  # type: ignore[override]
        self,
        value: typing.Any,
        attr: typing.Any,
        data: typing.Any,
        **kwargs: typing.Any,
    ) -> typing.Union[typing.Set[typing.Any], typing.FrozenSet[typing.Any], None]:
        optional_list = super()._deserialize(value, attr, data, **kwargs)
        return None if optional_list is None else self.set_type(optional_list)



import collections.abc
import numbers
from typing import (
    Any,
)

bytes_types = (bytes, bytearray)
integer_types = (int,)
text_types = (str,)
string_types = (bytes, str, bytearray)


def is_integer(value: Any) -> bool:
    return isinstance(value, integer_types) and not isinstance(value, bool)


def is_bytes(value: Any) -> bool:
    return isinstance(value, bytes_types)


def is_text(value: Any) -> bool:
    return isinstance(value, text_types)


def is_string(value: Any) -> bool:
    return isinstance(value, string_types)


def is_boolean(value: Any) -> bool:
    return isinstance(value, bool)


def is_dict(obj: Any) -> bool:
    return isinstance(obj, collections.abc.Mapping)


def is_list_like(obj: Any) -> bool:
    return not is_string(obj) and isinstance(obj, collections.abc.Sequence)


def is_list(obj: Any) -> bool:
    return isinstance(obj, list)


def is_tuple(obj: Any) -> bool:
    return isinstance(obj, tuple)


def is_null(obj: Any) -> bool:
    return obj is None


def is_number(obj: Any) -> bool:
    return isinstance(obj, numbers.Number)



from typing import List

from ..cairo.felt import CairoData


class OutOfBoundsError(Exception):
    def __init__(self, position: int, requested_size: int, remaining_size: int):
        super().__init__(
            f"Requested {requested_size} elements, {remaining_size} available."
        )
        self.position = position
        self.requested_size = requested_size
        self.remaining_len = remaining_size


class CalldataReader:
    _data: List[int]
    _position: int

    def __init__(self, data: List[int]):
        self._data = data
        self._position = 0

    @property
    def remaining_len(self) -> int:
        return len(self._data) - self._position

    def read(self, size: int) -> CairoData:
        if size < 1:
            raise ValueError("size must be greater than 0")

        if size > self.remaining_len:
            raise OutOfBoundsError(
                position=self._position,
                requested_size=size,
                remaining_size=self.remaining_len,
            )
        data = self._data[self._position : self._position + size]
        self._position += size
        return data



from typing import (
    AsyncGenerator,
    Optional,
    TypeVar,
)

import sys

if sys.version_info.minor >= 11:
    from typing import Literal, ParamSpec
else:
    from typing_extensions import Literal, ParamSpec

T = TypeVar("T")
P = ParamSpec("P")
K = TypeVar("K")
V = TypeVar("V")
TAsyncGenerator = TypeVar("TAsyncGenerator", bound=AsyncGenerator)
NumType = TypeVar("NumType", int, float)
HASH_BYTES = 32

# If more shared types start popping up here extract to types.py.
Endianness = Literal["big", "little"]
TComparable = TypeVar("TComparable", bound="Comparable")

def to_bytes(
    value: int,
    length: Optional[int] = None,
    byte_order: Optional[Endianness] = None,
    signed: Optional[bool] = None,
) -> bytes:
    """
    Converts the given integer to a bytes object of given length and byte order.
    The default values are 32B width (which is the hash result width) and 'big', respectively.
    """
    if length is None:
        length = HASH_BYTES

    if byte_order is None:
        byte_order = "big"

    if signed is None:
        signed = False

    return int.to_bytes(value, length=length, byteorder=byte_order, signed=signed)


def from_bytes(
    value: bytes,
    byte_order: Optional[Endianness] = None,
    signed: Optional[bool] = None,
) -> int:
    """
    Converts the given bytes object (parsed according to the given byte order) to an integer.
    Default byte order is 'big'.
    """
    if byte_order is None:
        byte_order = "big"

    if signed is None:
        signed = False

    return int.from_bytes(value, byteorder=byte_order, signed=signed)